# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from botocore import xform_name
from botocore.session import get_session

_KNOWN_REPLACEMENTS = [
    'whatsapp',
]


def _get_all_xform_operations():
    session = get_session()
    service_model = session.get_service_model("socialmessaging")
    transform_operations = []
    for operation in service_model.operation_names:
        for replacement in _KNOWN_REPLACEMENTS:
            if replacement in operation.lower():
                transform_operations.append((operation, replacement))
    return transform_operations


XFORM_OPERATIONS = _get_all_xform_operations()


@pytest.mark.validates_models
@pytest.mark.parametrize("operation, replacement", XFORM_OPERATIONS)
def test_known_replacements(operation, replacement):
    # Validates that if a replacement shows up in the lowercased version of an
    # operation, we will keep all of those characters together in the final operation
    # name
    assert replacement in xform_name(operation, '_')
    assert replacement in xform_name(operation, '-')



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.stub import Stubber
from tests import BaseSessionTest, ClientHTTPStubber


class TestRDSPresignUrlInjection(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.client = self.session.create_client('rds', 'us-west-2')
        self.http_stubber = ClientHTTPStubber(self.client)

    def assert_presigned_url_injected_in_request(self, body):
        self.assertIn('PreSignedUrl', body)
        self.assertNotIn('SourceRegion', body)

    def test_copy_snapshot(self):
        params = {
            'SourceDBSnapshotIdentifier': 'source-db',
            'TargetDBSnapshotIdentifier': 'target-db',
            'SourceRegion': 'us-east-1',
        }
        response_body = (
            b'<CopyDBSnapshotResponse>'
            b'<CopyDBSnapshotResult></CopyDBSnapshotResult>'
            b'</CopyDBSnapshotResponse>'
        )
        self.http_stubber.add_response(body=response_body)
        with self.http_stubber:
            self.client.copy_db_snapshot(**params)
            sent_request = self.http_stubber.requests[0]
            self.assert_presigned_url_injected_in_request(sent_request.body)

    def test_create_db_instance_read_replica(self):
        params = {
            'SourceDBInstanceIdentifier': 'source-db',
            'DBInstanceIdentifier': 'target-db',
            'SourceRegion': 'us-east-1',
        }
        response_body = (
            b'<CreateDBInstanceReadReplicaResponse>'
            b'<CreateDBInstanceReadReplicaResult>'
            b'</CreateDBInstanceReadReplicaResult>'
            b'</CreateDBInstanceReadReplicaResponse>'
        )
        self.http_stubber.add_response(body=response_body)
        with self.http_stubber:
            self.client.create_db_instance_read_replica(**params)
            sent_request = self.http_stubber.requests[0]
            self.assert_presigned_url_injected_in_request(sent_request.body)

    def test_start_db_instance_automated_backups_replication(self):
        params = {
            'SourceDBInstanceArn': 'arn:aws:rds:us-east-1:123456789012:db:source-db-instance',
            'SourceRegion': 'us-east-1',
        }
        response_body = (
            b'<StartDBInstanceAutomatedBackupsReplicationResponse>'
            b'<StartDBInstanceAutomatedBackupsReplicationResult>'
            b'</StartDBInstanceAutomatedBackupsReplicationResult>'
            b'</StartDBInstanceAutomatedBackupsReplicationResponse>'
        )
        self.http_stubber.add_response(body=response_body)
        with self.http_stubber:
            self.client.start_db_instance_automated_backups_replication(
                **params
            )
            sent_request = self.http_stubber.requests[0]
            self.assert_presigned_url_injected_in_request(sent_request.body)


class TestRDS(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.client = self.session.create_client('rds', 'us-west-2')
        self.stubber = Stubber(self.client)
        self.stubber.activate()

    def test_generate_db_auth_token(self):
        hostname = 'host.us-east-1.rds.amazonaws.com'
        port = 3306
        username = 'mySQLUser'
        auth_token = self.client.generate_db_auth_token(
            DBHostname=hostname, Port=port, DBUsername=username
        )

        endpoint_url = 'host.us-east-1.rds.amazonaws.com:3306'
        self.assertIn(endpoint_url, auth_token)
        self.assertIn('Action=connect', auth_token)

        # Asserts that there is no scheme in the url
        self.assertTrue(auth_token.startswith(hostname))



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import jmespath
import pytest
from jsonschema import Draft4Validator

import botocore.session
from botocore.exceptions import UnknownServiceError
from botocore.utils import ArgumentGenerator

WAITER_SCHEMA = {
    "type": "object",
    "properties": {
        "version": {"type": "number"},
        "waiters": {
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "properties": {
                    "type": {"type": "string", "enum": ["api"]},
                    "operation": {"type": "string"},
                    "description": {"type": "string"},
                    "delay": {
                        "type": "number",
                        "minimum": 0,
                    },
                    "maxAttempts": {"type": "integer", "minimum": 1},
                    "acceptors": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "state": {
                                    "type": "string",
                                    "enum": ["success", "retry", "failure"],
                                },
                                "matcher": {
                                    "type": "string",
                                    "enum": [
                                        "path",
                                        "pathAll",
                                        "pathAny",
                                        "status",
                                        "error",
                                    ],
                                },
                                "argument": {"type": "string"},
                                "expected": {
                                    "oneOf": [
                                        {"type": "string"},
                                        {"type": "number"},
                                        {"type": "boolean"},
                                    ]
                                },
                            },
                            "required": ["state", "matcher", "expected"],
                            "additionalProperties": False,
                        },
                    },
                },
                "required": ["operation", "delay", "maxAttempts", "acceptors"],
                "additionalProperties": False,
            },
        },
    },
    "additionalProperties": False,
}


def _waiter_configs():
    session = botocore.session.get_session()
    validator = Draft4Validator(WAITER_SCHEMA)
    for service_name in session.get_available_services():
        client = session.create_client(service_name, 'us-east-1')
        try:
            # We use the loader directly here because we need the entire
            # json document, not just the portions exposed (either
            # internally or externally) by the WaiterModel class.
            loader = session.get_component('data_loader')
            waiter_model = loader.load_service_model(service_name, 'waiters-2')
        except UnknownServiceError:
            # The service doesn't have waiters
            continue
        yield validator, waiter_model, client


@pytest.mark.parametrize("validator, waiter_model, client", _waiter_configs())
def test_lint_waiter_configs(validator, waiter_model, client):
    _validate_schema(validator, waiter_model)
    for waiter_name in client.waiter_names:
        _lint_single_waiter(client, waiter_name, client.meta.service_model)


def _lint_single_waiter(client, waiter_name, service_model):
    try:
        waiter = client.get_waiter(waiter_name)
        # The 'acceptors' property is dynamic and will create
        # the acceptor configs when first accessed.  This is still
        # considered a failure to construct the waiter which is
        # why it's in this try/except block.
        # This catches things like:
        # * jmespath expression compiles
        # * matcher has a known value
        acceptors = waiter.config.acceptors
    except Exception as e:
        raise AssertionError(f"Could not create waiter '{waiter_name}': {e}")
    operation_name = waiter.config.operation
    # Needs to reference an existing operation name.
    if operation_name not in service_model.operation_names:
        raise AssertionError(
            "Waiter config references unknown " f"operation: {operation_name}"
        )
    # Needs to have at least one acceptor.
    if not waiter.config.acceptors:
        raise AssertionError(
            "Waiter config must have at least "
            f"one acceptor state: {waiter.name}"
        )
    op_model = service_model.operation_model(operation_name)
    for acceptor in acceptors:
        _validate_acceptor(acceptor, op_model, waiter.name)

    if not waiter.name.isalnum():
        raise AssertionError(f"Waiter name {waiter_name} is not alphanumeric.")


def _validate_schema(validator, waiter_json):
    errors = list(e.message for e in validator.iter_errors(waiter_json))
    if errors:
        raise AssertionError('\n'.join(errors))


def _validate_acceptor(acceptor, op_model, waiter_name):
    if acceptor.matcher.startswith('path'):
        expression = acceptor.argument
        # The JMESPath expression should have the potential to match something
        # in the response shape.
        output_shape = op_model.output_shape
        assert (
            output_shape is not None
        ), f"Waiter '{waiter_name}' has JMESPath expression with no output shape: {op_model}"
        # We want to check if the JMESPath expression makes sense.
        # To do this, we'll generate sample output and evaluate the
        # JMESPath expression against the output.  We'll then
        # check a few things about this returned search result.
        search_result = _search_jmespath_expression(expression, op_model)
        if search_result is None:
            raise AssertionError(
                f"JMESPath expression did not match anything for waiter "
                f"'{waiter_name}': {expression}"
            )
        if acceptor.matcher in ['pathAll', 'pathAny']:
            assert isinstance(search_result, list), (
                f"Attempted to use '{acceptor.matcher}' matcher in waiter "
                f"'{waiter_name}' with non list result in JMESPath expression: "
                f"{expression}"
            )


def _search_jmespath_expression(expression, op_model):
    arg_gen = ArgumentGenerator(use_member_names=True)
    sample_output = arg_gen.generate_skeleton(op_model.output_shape)
    search_result = jmespath.search(expression, sample_output)
    return search_result



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import copy

from botocore.compat import OrderedDict
from botocore.endpoint import DEFAULT_TIMEOUT, MAX_POOL_CONNECTIONS
from botocore.exceptions import (
    InvalidMaxRetryAttemptsError,
    InvalidRetryConfigurationError,
    InvalidRetryModeError,
    InvalidS3AddressingStyleError,
)


class Config:
    """Advanced configuration for Botocore clients.

    :type region_name: str
    :param region_name: The region to use in instantiating the client

    :type signature_version: str
    :param signature_version: The signature version when signing requests.

    :type user_agent: str
    :param user_agent: The value to use in the User-Agent header.

    :type user_agent_extra: str
    :param user_agent_extra: The value to append to the current User-Agent
        header value.

    :type user_agent_appid: str
    :param user_agent_appid: A value that gets included in the User-Agent
        string in the format "app/<user_agent_appid>". Allowed characters are
        ASCII alphanumerics and ``!$%&'*+-.^_`|~``. All other characters will
        be replaced by a ``-``.

    :type connect_timeout: float or int
    :param connect_timeout: The time in seconds till a timeout exception is
        thrown when attempting to make a connection. The default is 60
        seconds.

    :type read_timeout: float or int
    :param read_timeout: The time in seconds till a timeout exception is
        thrown when attempting to read from a connection. The default is
        60 seconds.

    :type parameter_validation: bool
    :param parameter_validation: Whether parameter validation should occur
        when serializing requests. The default is True.  You can disable
        parameter validation for performance reasons.  Otherwise, it's
        recommended to leave parameter validation enabled.

    :type max_pool_connections: int
    :param max_pool_connections: The maximum number of connections to
        keep in a connection pool.  If this value is not set, the default
        value of 10 is used.

    :type proxies: dict
    :param proxies: A dictionary of proxy servers to use by protocol or
        endpoint, e.g.:
        ``{'http': 'foo.bar:3128', 'http://hostname': 'foo.bar:4012'}``.
        The proxies are used on each request.

    :type proxies_config: dict
    :param proxies_config: A dictionary of additional proxy configurations.
        Valid keys are:

        * ``proxy_ca_bundle`` -- The path to a custom certificate bundle to use
          when establishing SSL/TLS connections with proxy.

        * ``proxy_client_cert`` -- The path to a certificate for proxy
          TLS client authentication.

          When a string is provided it is treated as a path to a proxy client
          certificate. When a two element tuple is provided, it will be
          interpreted as the path to the client certificate, and the path
          to the certificate key.

        * ``proxy_use_forwarding_for_https`` -- For HTTPS proxies,
          forward your requests to HTTPS destinations with an absolute
          URI. We strongly recommend you only use this option with
          trusted or corporate proxies. Value must be boolean.

    :type s3: dict
    :param s3: A dictionary of S3 specific configurations.
        Valid keys are:

        * ``use_accelerate_endpoint`` -- Refers to whether to use the S3
          Accelerate endpoint. The value must be a boolean. If True, the
          client will use the S3 Accelerate endpoint. If the S3 Accelerate
          endpoint is being used then the addressing style will always
          be virtual.

        * ``payload_signing_enabled`` -- Refers to whether or not to SHA256
          sign sigv4 payloads. By default, this is disabled for streaming
          uploads (UploadPart and PutObject).

        * ``addressing_style`` -- Refers to the style in which to address
          s3 endpoints. Values must be a string that equals one of:

          * ``auto`` -- Addressing style is chosen for user. Depending
            on the configuration of client, the endpoint may be addressed in
            the virtual or the path style. Note that this is the default
            behavior if no style is specified.

          * ``virtual`` -- Addressing style is always virtual. The name of the
            bucket must be DNS compatible or an exception will be thrown.
            Endpoints will be addressed as such: ``amzn-s3-demo-bucket.s3.amazonaws.com``

          * ``path`` -- Addressing style is always by path. Endpoints will be
            addressed as such: ``s3.amazonaws.com/amzn-s3-demo-bucket``

        * ``us_east_1_regional_endpoint`` -- Refers to what S3 endpoint to use
          when the region is configured to be us-east-1. Values must be a
          string that equals:

          * ``regional`` -- Use the us-east-1.amazonaws.com endpoint if the
            client is configured to use the us-east-1 region.

          * ``legacy`` -- Use the s3.amazonaws.com endpoint if the client is
            configured to use the us-east-1 region. This is the default if
            the configuration option is not specified.


    :type retries: dict
    :param retries: A dictionary for configuration related to retry behavior.
        Valid keys are:

        * ``total_max_attempts`` -- An integer representing the maximum number of
          total attempts that will be made on a single request.  This includes
          the initial request, so a value of 1 indicates that no requests
          will be retried.  If ``total_max_attempts`` and ``max_attempts``
          are both provided, ``total_max_attempts`` takes precedence.
          ``total_max_attempts`` is preferred over ``max_attempts`` because
          it maps to the ``AWS_MAX_ATTEMPTS`` environment variable and
          the ``max_attempts`` config file value.
        * ``max_attempts`` -- An integer representing the maximum number of
          retry attempts that will be made on a single request. For
          example, setting this value to 2 will result in the request
          being retried at most two times after the initial request. Setting
          this value to 0 will result in no retries ever being attempted after
          the initial request. If not provided, the number of retries will
          default to the value specified in the service model, which is
          typically four retries.
        * ``mode`` -- A string representing the type of retry mode botocore
          should use.  Valid values are:

          * ``legacy`` - The pre-existing retry behavior.

          * ``standard`` - The standardized set of retry rules. This will also
            default to 3 max attempts unless overridden.

          * ``adaptive`` - Retries with additional client side throttling.

    :type client_cert: str, (str, str)
    :param client_cert: The path to a certificate for TLS client authentication.

        When a string is provided it is treated as a path to a client
        certificate to be used when creating a TLS connection.

        If a client key is to be provided alongside the client certificate the
        client_cert should be set to a tuple of length two where the first
        element is the path to the client certificate and the second element is
        the path to the certificate key.

    :type inject_host_prefix: bool
    :param inject_host_prefix: Whether host prefix injection should occur.

        Defaults to True.

        Setting this to False disables the injection of operation parameters
        into the prefix of the hostname. This is useful for clients providing
        custom endpoints that should not have their host prefix modified.

    :type use_dualstack_endpoint: bool
    :param use_dualstack_endpoint: Setting to True enables dualstack
        endpoint resolution.

        Defaults to None.

    :type use_fips_endpoint: bool
    :param use_fips_endpoint: Setting to True enables fips
        endpoint resolution.

        Defaults to None.

    :type ignore_configured_endpoint_urls: bool
    :param ignore_configured_endpoint_urls: Setting to True disables use
        of endpoint URLs provided via environment variables and
        the shared configuration file.

        Defaults to None.

    :type tcp_keepalive: bool
    :param tcp_keepalive: Enables the TCP Keep-Alive socket option used when
        creating new connections if set to True.

        Defaults to False.

    :type request_min_compression_size_bytes: int
    :param request_min_compression_size_bytes: The minimum size in bytes that a
        request body should be to trigger compression. All requests with
        streaming input that don't contain the ``requiresLength`` trait will be
        compressed regardless of this setting.

        Defaults to None.

    :type disable_request_compression: bool
    :param disable_request_compression: Disables request body compression if
        set to True.

        Defaults to None.

    :type sigv4a_signing_region_set: string
    :param sigv4a_signing_region_set: A set of AWS regions to apply the signature for
        when using SigV4a for signing. Set to ``*`` to represent all regions.

        Defaults to None.

    :type client_context_params: dict
    :param client_context_params: A dictionary of parameters specific to
        individual services. If available, valid parameters can be found in
        the ``Client Context Parameters`` section of the service client's
        documentation. Invalid parameters or ones that are not used by the
        specified service will be ignored.

        Defaults to None.
    """

    OPTION_DEFAULTS = OrderedDict(
        [
            ('region_name', None),
            ('signature_version', None),
            ('user_agent', None),
            ('user_agent_extra', None),
            ('user_agent_appid', None),
            ('connect_timeout', DEFAULT_TIMEOUT),
            ('read_timeout', DEFAULT_TIMEOUT),
            ('parameter_validation', True),
            ('max_pool_connections', MAX_POOL_CONNECTIONS),
            ('proxies', None),
            ('proxies_config', None),
            ('s3', None),
            ('retries', None),
            ('client_cert', None),
            ('inject_host_prefix', True),
            ('endpoint_discovery_enabled', None),
            ('use_dualstack_endpoint', None),
            ('use_fips_endpoint', None),
            ('ignore_configured_endpoint_urls', None),
            ('defaults_mode', None),
            ('tcp_keepalive', None),
            ('request_min_compression_size_bytes', None),
            ('disable_request_compression', None),
            ('client_context_params', None),
            ('sigv4a_signing_region_set', None),
        ]
    )

    NON_LEGACY_OPTION_DEFAULTS = {
        'connect_timeout': None,
    }

    def __init__(self, *args, **kwargs):
        self._user_provided_options = self._record_user_provided_options(
            args, kwargs
        )

        # Merge the user_provided options onto the default options
        config_vars = copy.copy(self.OPTION_DEFAULTS)
        defaults_mode = self._user_provided_options.get(
            'defaults_mode', 'legacy'
        )
        if defaults_mode != 'legacy':
            config_vars.update(self.NON_LEGACY_OPTION_DEFAULTS)
        config_vars.update(self._user_provided_options)

        # Set the attributes based on the config_vars
        for key, value in config_vars.items():
            setattr(self, key, value)

        # Validate the s3 options
        self._validate_s3_configuration(self.s3)

        self._validate_retry_configuration(self.retries)

    def _record_user_provided_options(self, args, kwargs):
        option_order = list(self.OPTION_DEFAULTS)
        user_provided_options = {}

        # Iterate through the kwargs passed through to the constructor and
        # map valid keys to the dictionary
        for key, value in kwargs.items():
            if key in self.OPTION_DEFAULTS:
                user_provided_options[key] = value
            # The key must exist in the available options
            else:
                raise TypeError(f"Got unexpected keyword argument '{key}'")

        # The number of args should not be longer than the allowed
        # options
        if len(args) > len(option_order):
            raise TypeError(
                f"Takes at most {len(option_order)} arguments ({len(args)} given)"
            )

        # Iterate through the args passed through to the constructor and map
        # them to appropriate keys.
        for i, arg in enumerate(args):
            # If a kwarg was specified for the arg, then error out
            if option_order[i] in user_provided_options:
                raise TypeError(
                    f"Got multiple values for keyword argument '{option_order[i]}'"
                )
            user_provided_options[option_order[i]] = arg

        return user_provided_options

    def _validate_s3_configuration(self, s3):
        if s3 is not None:
            addressing_style = s3.get('addressing_style')
            if addressing_style not in ['virtual', 'auto', 'path', None]:
                raise InvalidS3AddressingStyleError(
                    s3_addressing_style=addressing_style
                )

    def _validate_retry_configuration(self, retries):
        valid_options = ('max_attempts', 'mode', 'total_max_attempts')
        valid_modes = ('legacy', 'standard', 'adaptive')
        if retries is not None:
            for key, value in retries.items():
                if key not in valid_options:
                    raise InvalidRetryConfigurationError(
                        retry_config_option=key,
                        valid_options=valid_options,
                    )
                if key == 'max_attempts' and value < 0:
                    raise InvalidMaxRetryAttemptsError(
                        provided_max_attempts=value,
                        min_value=0,
                    )
                if key == 'total_max_attempts' and value < 1:
                    raise InvalidMaxRetryAttemptsError(
                        provided_max_attempts=value,
                        min_value=1,
                    )
                if key == 'mode' and value not in valid_modes:
                    raise InvalidRetryModeError(
                        provided_retry_mode=value,
                        valid_modes=valid_modes,
                    )

    def merge(self, other_config):
        """Merges the config object with another config object

        This will merge in all non-default values from the provided config
        and return a new config object

        :type other_config: botocore.config.Config
        :param other config: Another config object to merge with. The values
            in the provided config object will take precedence in the merging

        :returns: A config object built from the merged values of both
            config objects.
        """
        # Make a copy of the current attributes in the config object.
        config_options = copy.copy(self._user_provided_options)

        # Merge in the user provided options from the other config
        config_options.update(other_config._user_provided_options)

        # Return a new config object with the merged properties.
        return Config(**config_options)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from tests import ALL_SERVICES


def _all_service_error_shapes():
    for service_model in ALL_SERVICES:
        yield from service_model.error_shapes


def _all_operations():
    for service_model in ALL_SERVICES:
        for operation_name in service_model.operation_names:
            yield service_model.operation_model(operation_name).output_shape


def _assert_not_shadowed(key, shape):
    if not shape:
        return

    assert (
        key not in shape.members
    ), f'Found shape "{shape.name}" that shadows the botocore response key "{key}"'


@pytest.mark.parametrize("operation_output_shape", _all_operations())
def test_response_metadata_is_not_shadowed(operation_output_shape):
    _assert_not_shadowed('ResponseMetadata', operation_output_shape)


@pytest.mark.parametrize("error_shape", _all_service_error_shapes())
def test_exceptions_do_not_shadow_response_metadata(error_shape):
    _assert_not_shadowed('ResponseMetadata', error_shape)


@pytest.mark.parametrize("error_shape", _all_service_error_shapes())
def test_exceptions_do_not_shadow_error(error_shape):
    _assert_not_shadowed('Error', error_shape)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from tests import BaseSessionTest, ClientHTTPStubber


class TestNeptunePresignUrlInjection(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.client = self.session.create_client('neptune', 'us-west-2')
        self.http_stubber = ClientHTTPStubber(self.client)

    def assert_presigned_url_injected_in_request(self, body):
        self.assertIn('PreSignedUrl', body)
        self.assertNotIn('SourceRegion', body)

    def test_create_db_cluster(self):
        params = {
            'DBClusterIdentifier': 'my-cluster',
            'Engine': 'neptune',
            'SourceRegion': 'us-east-1',
        }
        response_body = (
            b'<CreateDBClusterResponse>'
            b'<CreateDBClusterResult>'
            b'</CreateDBClusterResult>'
            b'</CreateDBClusterResponse>'
        )
        self.http_stubber.add_response(body=response_body)
        with self.http_stubber:
            self.client.create_db_cluster(**params)
            sent_request = self.http_stubber.requests[0]
            self.assert_presigned_url_injected_in_request(sent_request.body)

    def test_copy_db_cluster_snapshot(self):
        params = {
            'SourceDBClusterSnapshotIdentifier': 'source-db',
            'TargetDBClusterSnapshotIdentifier': 'target-db',
            'SourceRegion': 'us-east-1',
        }
        response_body = (
            b'<CopyDBClusterSnapshotResponse>'
            b'<CopyDBClusterSnapshotResult>'
            b'</CopyDBClusterSnapshotResult>'
            b'</CopyDBClusterSnapshotResponse>'
        )
        self.http_stubber.add_response(body=response_body)
        with self.http_stubber:
            self.client.copy_db_cluster_snapshot(**params)
            sent_request = self.http_stubber.requests[0]
            self.assert_presigned_url_injected_in_request(sent_request.body)



from collections import Counter

import pytest

from botocore import model
from botocore.awsrequest import AWSResponse
from botocore.exceptions import (
    ConnectionError,
    HTTPClientError,
    ReadTimeoutError,
)
from botocore.retries import quota, standard
from tests import mock, unittest

RETRYABLE_THROTTLED_RESPONSES = [
    # From the spec under "Throttling Errors"
    # The status codes technically don't matter here, but we're adding
    # them for completeness.
    # StatusCode, ErrorCode, Retryable?
    (400, 'Throttling', True),
    (400, 'ThrottlingException', True),
    (400, 'ThrottledException', True),
    (400, 'RequestThrottledException', True),
    (400, 'TooManyRequestsException', True),
    (400, 'ProvisionedThroughputExceededException', True),
    (400, 'TransactionInProgressException', True),
    (503, 'RequestLimitExceeded', True),
    (509, 'BandwidthLimitExceeded', True),
    (400, 'LimitExceededException', True),
    (403, 'RequestThrottled', True),
    (503, 'SlowDown', True),
    (400, 'PriorRequestNotComplete', True),
    (502, 'EC2ThrottledException', True),
    # These are some negative test cases, not in the spec but we'll use
    # to verify we can detect throttled errors correctly.
    (400, 'NotAThrottlingError', False),
    (500, 'InternalServerError', False),
    # "None" here represents no parsed response we just have a plain
    # HTTP response and a 400 status code response.
    (400, None, False),
    (500, None, False),
    (200, None, False),
]

RETRYABLE_TRANSIENT_ERRORS = [
    # StatusCode, Error, Retryable?
    (400, 'RequestTimeout', True),
    (400, 'RequestTimeoutException', True),
    (400, 'PriorRequestNotComplete', True),
    # "Any HTTP response with an HTTP status code of 500, 502, 503, or 504".
    (500, None, True),
    (502, None, True),
    (503, None, True),
    (504, None, True),
    # We'll also add a few errors with an explicit error code to verify
    # that the code doesn't matter.
    (500, 'InternalServiceError', True),
    (502, 'BadError', True),
    # These are botocore specific errors that correspond to
    # "Any IO (socket) level error where we are unable to read an HTTP
    # response.
    (None, ConnectionError(error='unknown'), True),
    (None, HTTPClientError(error='unknown'), True),
    # Negative cases
    (200, None, False),
    # This is a throttling error not a transient error
    (400, 'Throttling', False),
    (400, None, False),
]


# These tests are intended to be paired with the
# SERVICE_DESCRIPTION_WITH_RETRIES definition.
RETRYABLE_MODELED_ERRORS = [
    (400, 'ModeledThrottlingError', True),
    (400, 'ModeledRetryableError', True),
    # Note this is ErrorCodeRetryable, not ModeledRetryableErrorWithCode,
    # because the shape has a error code defined for it.
    (400, 'ErrorCodeRetryable', True),
    (400, 'NonRetryableError', False),
    (None, ConnectionError(error='unknown'), False),
]


SERVICE_DESCRIPTION_WITH_RETRIES = {
    'metadata': {},
    'operations': {
        'TestOperation': {
            'name': 'TestOperation',
            'input': {'shape': 'FakeInputOutputShape'},
            'output': {'shape': 'FakeInputOutputShape'},
            'errors': [
                {'shape': 'ModeledThrottlingError'},
                {'shape': 'ModeledRetryableError'},
                {'shape': 'ModeledRetryableErrorWithCode'},
                {'shape': 'NonRetryableError'},
            ],
        }
    },
    'shapes': {
        'FakeInputOutputShape': {
            'type': 'structure',
            'members': {},
        },
        'ModeledThrottlingError': {
            'type': 'structure',
            'members': {
                'message': {
                    'shape': 'ErrorMessage',
                }
            },
            'exception': True,
            'retryable': {'throttling': True},
        },
        'ModeledRetryableError': {
            'type': 'structure',
            'members': {
                'message': {
                    'shape': 'ErrorMessage',
                }
            },
            'exception': True,
            'retryable': {},
        },
        'ModeledRetryableErrorWithCode': {
            'type': 'structure',
            'members': {
                'message': {
                    'shape': 'ErrorMessage',
                }
            },
            'error': {
                'code': 'ErrorCodeRetryable',
            },
            'exception': True,
            'retryable': {'throttling': True},
        },
        'NonRetryableError': {
            'type': 'structure',
            'members': {
                'message': {
                    'shape': 'ErrorMessage',
                }
            },
            'exception': True,
        },
    },
}


@pytest.mark.parametrize('case', RETRYABLE_TRANSIENT_ERRORS)
def test_can_detect_retryable_transient_errors(case):
    transient_checker = standard.TransientRetryableChecker()
    _verify_retryable(transient_checker, None, *case)


@pytest.mark.parametrize('case', RETRYABLE_THROTTLED_RESPONSES)
def test_can_detect_retryable_throttled_errors(case):
    throttled_checker = standard.ThrottledRetryableChecker()
    _verify_retryable(throttled_checker, None, *case)


@pytest.mark.parametrize('case', RETRYABLE_MODELED_ERRORS)
def test_can_detect_modeled_retryable_errors(case):
    modeled_retry_checker = standard.ModeledRetryableChecker()
    _verify_retryable(
        modeled_retry_checker, get_operation_model_with_retries(), *case
    )


@pytest.mark.parametrize(
    'case',
    [
        case
        for case in RETRYABLE_TRANSIENT_ERRORS
        + RETRYABLE_THROTTLED_RESPONSES
        + RETRYABLE_MODELED_ERRORS
        if case[2]
    ],
)
def test_standard_retry_conditions(case):
    """This is verifying that the high level object used for checking
    retry conditions still handles all the individual testcases.

    It's possible that cases that are retryable for an individual checker
    aren't retryable for a different checker.  We need to filter out all
    the False cases (if case[2]).
    """
    standard_checker = standard.StandardRetryConditions()
    op_model = get_operation_model_with_retries()
    _verify_retryable(standard_checker, op_model, *case)


def get_operation_model_with_retries():
    service = model.ServiceModel(
        SERVICE_DESCRIPTION_WITH_RETRIES, service_name='my-service'
    )
    return service.operation_model('TestOperation')


def _verify_retryable(
    checker, operation_model, status_code, error, is_retryable
):
    http_response = AWSResponse(
        status_code=status_code, raw=None, headers={}, url='https://foo/'
    )
    parsed_response = None
    caught_exception = None
    if error is not None:
        if isinstance(error, Exception):
            caught_exception = error
        else:
            parsed_response = {'Error': {'Code': error, 'Message': 'Error'}}
    context = standard.RetryContext(
        attempt_number=1,
        operation_model=operation_model,
        parsed_response=parsed_response,
        http_response=http_response,
        caught_exception=caught_exception,
    )
    assert checker.is_retryable(context) == is_retryable


def arbitrary_retry_context():
    # Used when you just need a dummy retry context that looks like
    # a failed request.
    return standard.RetryContext(
        attempt_number=1,
        operation_model=None,
        parsed_response={'Error': {'Code': 'ErrorCode', 'Message': 'message'}},
        http_response=AWSResponse(
            status_code=500, raw=None, headers={}, url='https://foo'
        ),
        caught_exception=None,
    )


def test_can_honor_max_attempts():
    checker = standard.MaxAttemptsChecker(max_attempts=3)
    context = arbitrary_retry_context()
    context.attempt_number = 1
    assert checker.is_retryable(context) is True

    context.attempt_number = 2
    assert checker.is_retryable(context) is True

    context.attempt_number = 3
    assert checker.is_retryable(context) is False


def test_max_attempts_adds_metadata_key_when_reached():
    checker = standard.MaxAttemptsChecker(max_attempts=3)
    context = arbitrary_retry_context()
    context.attempt_number = 3
    assert checker.is_retryable(context) is False
    assert context.get_retry_metadata() == {'MaxAttemptsReached': True}


def test_retries_context_not_on_request_context():
    checker = standard.MaxAttemptsChecker(max_attempts=3)
    context = arbitrary_retry_context()
    context.attempt_number = 3
    assert checker.is_retryable(context) is False
    assert context.request_context == {}


def test_can_create_default_retry_handler():
    mock_client = mock.Mock()
    mock_client.meta.service_model.service_id = model.ServiceId('my-service')
    assert isinstance(
        standard.register_retry_handler(mock_client), standard.RetryHandler
    )
    call_args_list = mock_client.meta.events.register.call_args_list
    # We should have registered the retry quota to after-calls
    first_call = call_args_list[0][0]
    second_call = call_args_list[1][0]
    # Not sure if there's a way to verify the class associated with the
    # bound method matches what we expect.
    assert first_call[0] == 'after-call.my-service'
    assert second_call[0] == 'needs-retry.my-service'


class TestRetryHandler(unittest.TestCase):
    def setUp(self):
        self.retry_policy = mock.Mock(spec=standard.RetryPolicy)
        self.retry_event_adapter = mock.Mock(spec=standard.RetryEventAdapter)
        self.retry_quota = mock.Mock(spec=standard.RetryQuotaChecker)
        self.retry_handler = standard.RetryHandler(
            retry_policy=self.retry_policy,
            retry_event_adapter=self.retry_event_adapter,
            retry_quota=self.retry_quota,
        )

    def test_does_need_retry(self):
        self.retry_event_adapter.create_retry_context.return_value = (
            mock.sentinel.retry_context
        )
        self.retry_policy.should_retry.return_value = True
        self.retry_quota.acquire_retry_quota.return_value = True
        self.retry_policy.compute_retry_delay.return_value = 1

        self.assertEqual(self.retry_handler.needs_retry(fake_kwargs='foo'), 1)
        self.retry_event_adapter.create_retry_context.assert_called_with(
            fake_kwargs='foo'
        )
        self.retry_policy.should_retry.assert_called_with(
            mock.sentinel.retry_context
        )
        self.retry_quota.acquire_retry_quota.assert_called_with(
            mock.sentinel.retry_context
        )
        self.retry_policy.compute_retry_delay.assert_called_with(
            mock.sentinel.retry_context
        )

    def test_does_not_need_retry(self):
        self.retry_event_adapter.create_retry_context.return_value = (
            mock.sentinel.retry_context
        )
        self.retry_policy.should_retry.return_value = False

        self.assertIsNone(self.retry_handler.needs_retry(fake_kwargs='foo'))
        # Shouldn't consult quota if we don't have a retryable condition.
        self.assertFalse(self.retry_quota.acquire_retry_quota.called)

    def test_needs_retry_but_not_enough_quota(self):
        self.retry_event_adapter.create_retry_context.return_value = (
            mock.sentinel.retry_context
        )
        self.retry_policy.should_retry.return_value = True
        self.retry_quota.acquire_retry_quota.return_value = False

        self.assertIsNone(self.retry_handler.needs_retry(fake_kwargs='foo'))

    def test_retry_handler_adds_retry_metadata_to_response(self):
        self.retry_event_adapter.create_retry_context.return_value = (
            mock.sentinel.retry_context
        )
        self.retry_policy.should_retry.return_value = False
        self.assertIsNone(self.retry_handler.needs_retry(fake_kwargs='foo'))
        adapter = self.retry_event_adapter
        adapter.adapt_retry_response_from_context.assert_called_with(
            mock.sentinel.retry_context
        )


class TestRetryEventAdapter(unittest.TestCase):
    def setUp(self):
        self.success_response = {'ResponseMetadata': {}, 'Foo': {}}
        self.failed_response = {'ResponseMetadata': {}, 'Error': {}}
        self.http_success = AWSResponse(
            status_code=200, raw=None, headers={}, url='https://foo/'
        )
        self.http_failed = AWSResponse(
            status_code=500, raw=None, headers={}, url='https://foo/'
        )
        self.caught_exception = ConnectionError(error='unknown')

    def test_create_context_from_success_response(self):
        context = standard.RetryEventAdapter().create_retry_context(
            response=(self.http_success, self.success_response),
            attempts=1,
            caught_exception=None,
            request_dict={'context': {'foo': 'bar'}},
            operation=mock.sentinel.operation_model,
        )

        self.assertEqual(context.attempt_number, 1)
        self.assertEqual(
            context.operation_model, mock.sentinel.operation_model
        )
        self.assertEqual(context.parsed_response, self.success_response)
        self.assertEqual(context.http_response, self.http_success)
        self.assertEqual(context.caught_exception, None)
        self.assertEqual(context.request_context, {'foo': 'bar'})

    def test_create_context_from_service_error(self):
        context = standard.RetryEventAdapter().create_retry_context(
            response=(self.http_failed, self.failed_response),
            attempts=1,
            caught_exception=None,
            request_dict={'context': {'foo': 'bar'}},
            operation=mock.sentinel.operation_model,
        )
        # We already tested the other attributes in
        # test_create_context_from_success_response so we're only checking
        # the attributes relevant to this test.
        self.assertEqual(context.parsed_response, self.failed_response)
        self.assertEqual(context.http_response, self.http_failed)

    def test_create_context_from_exception(self):
        context = standard.RetryEventAdapter().create_retry_context(
            response=None,
            attempts=1,
            caught_exception=self.caught_exception,
            request_dict={'context': {'foo': 'bar'}},
            operation=mock.sentinel.operation_model,
        )
        self.assertEqual(context.parsed_response, None)
        self.assertEqual(context.http_response, None)
        self.assertEqual(context.caught_exception, self.caught_exception)

    def test_can_inject_metadata_back_to_context(self):
        adapter = standard.RetryEventAdapter()
        context = adapter.create_retry_context(
            attempts=1,
            operation=None,
            caught_exception=None,
            request_dict={'context': {}},
            response=(self.http_failed, self.failed_response),
        )
        context.add_retry_metadata(MaxAttemptsReached=True)
        adapter.adapt_retry_response_from_context(context)
        self.assertEqual(
            self.failed_response['ResponseMetadata']['MaxAttemptsReached'],
            True,
        )


class TestRetryPolicy(unittest.TestCase):
    def setUp(self):
        self.retry_checker = mock.Mock(spec=standard.StandardRetryConditions)
        self.retry_backoff = mock.Mock(spec=standard.ExponentialBackoff)
        self.retry_policy = standard.RetryPolicy(
            retry_checker=self.retry_checker, retry_backoff=self.retry_backoff
        )

    def test_delegates_to_retry_checker(self):
        self.retry_checker.is_retryable.return_value = True
        self.assertTrue(self.retry_policy.should_retry(mock.sentinel.context))
        self.retry_checker.is_retryable.assert_called_with(
            mock.sentinel.context
        )

    def test_delegates_to_retry_backoff(self):
        self.retry_backoff.delay_amount.return_value = 1
        self.assertEqual(
            self.retry_policy.compute_retry_delay(mock.sentinel.context), 1
        )
        self.retry_backoff.delay_amount.assert_called_with(
            mock.sentinel.context
        )


class TestExponentialBackoff(unittest.TestCase):
    def setUp(self):
        self.random = lambda: 1
        self.backoff = standard.ExponentialBackoff(
            max_backoff=20, random=self.random
        )

    def test_range_of_exponential_backoff(self):
        backoffs = [
            self.backoff.delay_amount(standard.RetryContext(attempt_number=i))
            for i in range(1, 10)
        ]
        # Note that we're capped at 20 which is our max backoff.
        self.assertEqual(backoffs, [1, 2, 4, 8, 16, 20, 20, 20, 20])

    def test_exponential_backoff_with_jitter(self):
        backoff = standard.ExponentialBackoff()
        backoffs = [
            backoff.delay_amount(standard.RetryContext(attempt_number=3))
            for i in range(10)
        ]
        # For attempt number 3, we should have a max value of 4 (2 ^ 2),
        # so we can assert all the backoff values are within that range.
        for x in backoffs:
            self.assertTrue(0 <= x <= 4)

    def test_uniform_rand_dist_on_max_attempts(self):
        backoff = standard.ExponentialBackoff()
        num_datapoints = 10_000
        backoffs = [
            backoff.delay_amount(standard.RetryContext(attempt_number=10))
            for i in range(num_datapoints)
        ]
        self._assert_looks_like_uniform_distribution(backoffs)

    def _assert_looks_like_uniform_distribution(self, backoffs):
        histogram = Counter(int(el) for el in backoffs)
        expected_value = len(backoffs) / len(histogram)
        # This is an arbitrarily chosen tolerance, but we're being fairly
        # lenient here and giving a 20% tolerance.  We're only interested
        # in cases where it's obviously broken and not a uniform distribution.
        tolerance = 0.20
        low = expected_value - (expected_value * tolerance)
        high = expected_value + (expected_value * tolerance)
        out_of_range = [
            str(i) for i in histogram.values() if not low <= i <= high
        ]
        if out_of_range:
            raise AssertionError(
                "Backoff values outside of uniform distribution range "
                f"({low} - {high}): {', '.join(out_of_range)}"
            )


class TestRetryQuotaChecker(unittest.TestCase):
    def setUp(self):
        self.quota = quota.RetryQuota(500)
        self.quota_checker = standard.RetryQuotaChecker(self.quota)
        self.request_context = {}

    def create_context(self, is_timeout_error=False, status_code=200):
        caught_exception = None
        if is_timeout_error:
            caught_exception = ReadTimeoutError(endpoint_url='https://foo')
        http_response = AWSResponse(
            status_code=status_code, raw=None, headers={}, url='https://foo/'
        )
        context = standard.RetryContext(
            attempt_number=1,
            request_context=self.request_context,
            caught_exception=caught_exception,
            http_response=http_response,
        )
        return context

    def test_can_acquire_quota_non_timeout_error(self):
        self.assertTrue(
            self.quota_checker.acquire_retry_quota(self.create_context())
        )
        self.assertEqual(self.request_context['retry_quota_capacity'], 5)

    def test_can_acquire_quota_for_timeout_error(self):
        self.assertTrue(
            self.quota_checker.acquire_retry_quota(
                self.create_context(is_timeout_error=True)
            )
        )
        self.assertEqual(self.request_context['retry_quota_capacity'], 10)

    def test_can_release_quota_based_on_context_value_on_success(self):
        context = self.create_context()
        # This is where we had to retry the request but eventually
        # succeeded.
        http_response = self.create_context(status_code=200).http_response
        self.assertTrue(self.quota_checker.acquire_retry_quota(context))
        self.assertEqual(self.quota.available_capacity, 495)
        self.quota_checker.release_retry_quota(
            context.request_context, http_response=http_response
        )
        self.assertEqual(self.quota.available_capacity, 500)

    def test_dont_release_quota_if_all_retries_failed(self):
        context = self.create_context()
        # If max_attempts_reached is True, then it means we used up all
        # our retry attempts and still failed.  In this case we shouldn't
        # give any retry quota back.
        http_response = self.create_context(status_code=500).http_response
        self.assertTrue(self.quota_checker.acquire_retry_quota(context))
        self.assertEqual(self.quota.available_capacity, 495)
        self.quota_checker.release_retry_quota(
            context.request_context, http_response=http_response
        )
        self.assertEqual(self.quota.available_capacity, 495)

    def test_can_release_default_quota_if_not_in_context(self):
        context = self.create_context()
        self.assertTrue(self.quota_checker.acquire_retry_quota(context))
        self.assertEqual(self.quota.available_capacity, 495)
        # We're going to remove the quota amount from the request context.
        # This represents a successful request with no retries.
        self.request_context.pop('retry_quota_capacity')
        self.quota_checker.release_retry_quota(
            context.request_context, context.http_response
        )
        # We expect only 1 unit was released.
        self.assertEqual(self.quota.available_capacity, 496)

    def test_acquire_quota_fails(self):
        quota_checker = standard.RetryQuotaChecker(
            quota.RetryQuota(initial_capacity=5)
        )
        # The first one succeeds.
        self.assertTrue(
            quota_checker.acquire_retry_quota(self.create_context())
        )
        # But we should fail now because we're out of quota.
        self.request_context.pop('retry_quota_capacity')
        self.assertFalse(
            quota_checker.acquire_retry_quota(self.create_context())
        )
        self.assertNotIn('retry_quota_capacity', self.request_context)

    def test_quota_reached_adds_retry_metadata(self):
        quota_checker = standard.RetryQuotaChecker(
            quota.RetryQuota(initial_capacity=0)
        )
        context = self.create_context()
        self.assertFalse(quota_checker.acquire_retry_quota(context))
        self.assertEqual(
            context.get_retry_metadata(), {'RetryQuotaReached': True}
        )

    def test_single_failed_request_does_not_give_back_quota(self):
        context = self.create_context()
        http_response = self.create_context(status_code=400).http_response
        # First deduct some amount of the retry quota so we're not hitting
        # the upper bound.
        self.quota.acquire(50)
        self.assertEqual(self.quota.available_capacity, 450)
        self.quota_checker.release_retry_quota(
            context.request_context, http_response=http_response
        )
        self.assertEqual(self.quota.available_capacity, 450)


class TestRetryContext(unittest.TestCase):
    def test_can_get_error_code(self):
        context = arbitrary_retry_context()
        context.parsed_response['Error']['Code'] = 'MyErrorCode'
        self.assertEqual(context.get_error_code(), 'MyErrorCode')

    def test_no_error_code_if_no_parsed_response(self):
        context = arbitrary_retry_context()
        context.parsed_response = None
        self.assertIsNone(context.get_error_code())

    def test_no_error_code_returns_none(self):
        context = arbitrary_retry_context()
        context.parsed_response = {}
        self.assertIsNone(context.get_error_code())

    def test_can_add_retry_reason(self):
        context = arbitrary_retry_context()
        context.add_retry_metadata(MaxAttemptsReached=True)
        self.assertEqual(
            context.get_retry_metadata(), {'MaxAttemptsReached': True}
        )

    def test_handles_non_error_top_level_error_key_get_error_code(self):
        response = AWSResponse(
            status_code=200,
            raw=None,
            headers={},
            url='https://foo',
        )
        # A normal response can have a top level "Error" key that doesn't map
        # to an error code and should be ignored
        context = standard.RetryContext(
            attempt_number=1,
            operation_model=None,
            parsed_response={'Error': 'This is a 200 response body'},
            http_response=response,
            caught_exception=None,
        )
        self.assertEqual(context.get_error_code(), None)


class TestThrottlingErrorDetector(unittest.TestCase):
    def setUp(self):
        self.throttling_detector = standard.ThrottlingErrorDetector(
            standard.RetryEventAdapter()
        )

    def create_needs_retry_kwargs(self, **kwargs):
        retry_kwargs = {
            'response': None,
            'attempts': 1,
            'operation': None,
            'caught_exception': None,
            'request_dict': {'context': {}},
        }
        retry_kwargs.update(kwargs)
        return retry_kwargs

    def test_can_check_error_from_code(self):
        kwargs = self.create_needs_retry_kwargs()
        kwargs['response'] = (None, {'Error': {'Code': 'ThrottledException'}})
        self.assertTrue(self.throttling_detector.is_throttling_error(**kwargs))

    def test_no_throttling_error(self):
        kwargs = self.create_needs_retry_kwargs()
        kwargs['response'] = (None, {'Error': {'Code': 'RandomError'}})
        self.assertFalse(
            self.throttling_detector.is_throttling_error(**kwargs)
        )

    def test_detects_modeled_errors(self):
        kwargs = self.create_needs_retry_kwargs()
        kwargs['response'] = (
            None,
            {'Error': {'Code': 'ModeledThrottlingError'}},
        )
        kwargs['operation'] = get_operation_model_with_retries()
        self.assertTrue(self.throttling_detector.is_throttling_error(**kwargs))


class TestModeledRetryErrorDetector(unittest.TestCase):
    def setUp(self):
        self.modeled_error = standard.ModeledRetryErrorDetector()

    def test_not_retryable(self):
        context = arbitrary_retry_context()
        self.assertIsNone(self.modeled_error.detect_error_type(context))

    def test_transient_error(self):
        context = arbitrary_retry_context()
        context.parsed_response['Error']['Code'] = 'ModeledRetryableError'
        context.operation_model = get_operation_model_with_retries()
        self.assertEqual(
            self.modeled_error.detect_error_type(context),
            self.modeled_error.TRANSIENT_ERROR,
        )

    def test_throttling_error(self):
        context = arbitrary_retry_context()
        context.parsed_response['Error']['Code'] = 'ModeledThrottlingError'
        context.operation_model = get_operation_model_with_retries()
        self.assertEqual(
            self.modeled_error.detect_error_type(context),
            self.modeled_error.THROTTLING_ERROR,
        )


class Yes(standard.BaseRetryableChecker):
    def is_retryable(self, context):
        return True


class No(standard.BaseRetryableChecker):
    def is_retryable(self, context):
        return False


class TestOrRetryChecker(unittest.TestCase):
    def test_can_match_any_checker(self):
        self.assertTrue(standard.OrRetryChecker([Yes(), No()]))
        self.assertTrue(standard.OrRetryChecker([No(), Yes()]))
        self.assertTrue(standard.OrRetryChecker([Yes(), Yes()]))

    def test_false_if_no_checkers_match(self):
        self.assertTrue(standard.OrRetryChecker([No(), No(), No()]))



import pytest

from botocore.session import Session

# The list of services which were available when we switched over from using
# endpoint prefix in event to using service id. These should all accept
# either.
SERVICES = {
    "acm": {"endpoint_prefix": "acm", "service_id": "acm"},
    "acm-pca": {"endpoint_prefix": "acm-pca", "service_id": "acm-pca"},
    "apigateway": {
        "endpoint_prefix": "apigateway",
        "service_id": "api-gateway",
    },
    "application-autoscaling": {"service_id": "application-auto-scaling"},
    "appstream": {"endpoint_prefix": "appstream2", "service_id": "appstream"},
    "appsync": {"endpoint_prefix": "appsync", "service_id": "appsync"},
    "athena": {"endpoint_prefix": "athena", "service_id": "athena"},
    "autoscaling": {
        "endpoint_prefix": "autoscaling",
        "service_id": "auto-scaling",
    },
    "autoscaling-plans": {"service_id": "auto-scaling-plans"},
    "batch": {"endpoint_prefix": "batch", "service_id": "batch"},
    "budgets": {"endpoint_prefix": "budgets", "service_id": "budgets"},
    "ce": {"endpoint_prefix": "ce", "service_id": "cost-explorer"},
    "cloud9": {"endpoint_prefix": "cloud9", "service_id": "cloud9"},
    "clouddirectory": {
        "endpoint_prefix": "clouddirectory",
        "service_id": "clouddirectory",
    },
    "cloudformation": {
        "endpoint_prefix": "cloudformation",
        "service_id": "cloudformation",
    },
    "cloudfront": {
        "endpoint_prefix": "cloudfront",
        "service_id": "cloudfront",
    },
    "cloudhsm": {"endpoint_prefix": "cloudhsm", "service_id": "cloudhsm"},
    "cloudhsmv2": {
        "endpoint_prefix": "cloudhsmv2",
        "service_id": "cloudhsm-v2",
    },
    "cloudsearch": {
        "endpoint_prefix": "cloudsearch",
        "service_id": "cloudsearch",
    },
    "cloudsearchdomain": {
        "endpoint_prefix": "cloudsearchdomain",
        "service_id": "cloudsearch-domain",
    },
    "cloudtrail": {
        "endpoint_prefix": "cloudtrail",
        "service_id": "cloudtrail",
    },
    "cloudwatch": {
        "endpoint_prefix": "monitoring",
        "service_id": "cloudwatch",
    },
    "codebuild": {"endpoint_prefix": "codebuild", "service_id": "codebuild"},
    "codecommit": {
        "endpoint_prefix": "codecommit",
        "service_id": "codecommit",
    },
    "codedeploy": {
        "endpoint_prefix": "codedeploy",
        "service_id": "codedeploy",
    },
    "codepipeline": {
        "endpoint_prefix": "codepipeline",
        "service_id": "codepipeline",
    },
    "cognito-identity": {
        "endpoint_prefix": "cognito-identity",
        "service_id": "cognito-identity",
    },
    "cognito-idp": {
        "endpoint_prefix": "cognito-idp",
        "service_id": "cognito-identity-provider",
    },
    "cognito-sync": {
        "endpoint_prefix": "cognito-sync",
        "service_id": "cognito-sync",
    },
    "comprehend": {
        "endpoint_prefix": "comprehend",
        "service_id": "comprehend",
    },
    "config": {"endpoint_prefix": "config", "service_id": "config-service"},
    "connect": {"endpoint_prefix": "connect", "service_id": "connect"},
    "cur": {
        "endpoint_prefix": "cur",
        "service_id": "cost-and-usage-report-service",
    },
    "datapipeline": {
        "endpoint_prefix": "datapipeline",
        "service_id": "data-pipeline",
    },
    "dax": {"endpoint_prefix": "dax", "service_id": "dax"},
    "devicefarm": {
        "endpoint_prefix": "devicefarm",
        "service_id": "device-farm",
    },
    "directconnect": {
        "endpoint_prefix": "directconnect",
        "service_id": "direct-connect",
    },
    "discovery": {
        "endpoint_prefix": "discovery",
        "service_id": "application-discovery-service",
    },
    "dlm": {"endpoint_prefix": "dlm", "service_id": "dlm"},
    "dms": {
        "endpoint_prefix": "dms",
        "service_id": "database-migration-service",
    },
    "ds": {"endpoint_prefix": "ds", "service_id": "directory-service"},
    "dynamodb": {"endpoint_prefix": "dynamodb", "service_id": "dynamodb"},
    "dynamodbstreams": {
        "endpoint_prefix": "streams.dynamodb",
        "service_id": "dynamodb-streams",
    },
    "ec2": {"endpoint_prefix": "ec2", "service_id": "ec2"},
    "ecr": {"endpoint_prefix": "ecr", "service_id": "ecr"},
    "ecs": {"endpoint_prefix": "ecs", "service_id": "ecs"},
    "efs": {"endpoint_prefix": "elasticfilesystem", "service_id": "efs"},
    "eks": {"endpoint_prefix": "eks", "service_id": "eks"},
    "elasticache": {
        "endpoint_prefix": "elasticache",
        "service_id": "elasticache",
    },
    "elasticbeanstalk": {
        "endpoint_prefix": "elasticbeanstalk",
        "service_id": "elastic-beanstalk",
    },
    "elastictranscoder": {
        "endpoint_prefix": "elastictranscoder",
        "service_id": "elastic-transcoder",
    },
    "elb": {
        "endpoint_prefix": "elasticloadbalancing",
        "service_id": "elastic-load-balancing",
    },
    "elbv2": {"service_id": "elastic-load-balancing-v2"},
    "emr": {"endpoint_prefix": "elasticmapreduce", "service_id": "emr"},
    "es": {"endpoint_prefix": "es", "service_id": "elasticsearch-service"},
    "events": {"endpoint_prefix": "events", "service_id": "cloudwatch-events"},
    "firehose": {"endpoint_prefix": "firehose", "service_id": "firehose"},
    "fms": {"endpoint_prefix": "fms", "service_id": "fms"},
    "gamelift": {"endpoint_prefix": "gamelift", "service_id": "gamelift"},
    "glacier": {"endpoint_prefix": "glacier", "service_id": "glacier"},
    "glue": {"endpoint_prefix": "glue", "service_id": "glue"},
    "greengrass": {
        "endpoint_prefix": "greengrass",
        "service_id": "greengrass",
    },
    "guardduty": {"endpoint_prefix": "guardduty", "service_id": "guardduty"},
    "health": {"endpoint_prefix": "health", "service_id": "health"},
    "iam": {"endpoint_prefix": "iam", "service_id": "iam"},
    "importexport": {
        "endpoint_prefix": "importexport",
        "service_id": "importexport",
    },
    "inspector": {"endpoint_prefix": "inspector", "service_id": "inspector"},
    "iot": {"endpoint_prefix": "iot", "service_id": "iot"},
    "iot-data": {
        "endpoint_prefix": "data.iot",
        "service_id": "iot-data-plane",
    },
    "iot-jobs-data": {
        "endpoint_prefix": "data.jobs.iot",
        "service_id": "iot-jobs-data-plane",
    },
    "iot1click-devices": {
        "endpoint_prefix": "devices.iot1click",
        "service_id": "iot-1click-devices-service",
    },
    "iot1click-projects": {
        "endpoint_prefix": "projects.iot1click",
        "service_id": "iot-1click-projects",
    },
    "iotanalytics": {
        "endpoint_prefix": "iotanalytics",
        "service_id": "iotanalytics",
    },
    "kinesis": {"endpoint_prefix": "kinesis", "service_id": "kinesis"},
    "kinesis-video-archived-media": {
        "service_id": "kinesis-video-archived-media"
    },
    "kinesis-video-media": {"service_id": "kinesis-video-media"},
    "kinesisanalytics": {
        "endpoint_prefix": "kinesisanalytics",
        "service_id": "kinesis-analytics",
    },
    "kinesisvideo": {
        "endpoint_prefix": "kinesisvideo",
        "service_id": "kinesis-video",
    },
    "kms": {"endpoint_prefix": "kms", "service_id": "kms"},
    "lambda": {"endpoint_prefix": "lambda", "service_id": "lambda"},
    "lex-models": {
        "endpoint_prefix": "models.lex",
        "service_id": "lex-model-building-service",
    },
    "lex-runtime": {
        "endpoint_prefix": "runtime.lex",
        "service_id": "lex-runtime-service",
    },
    "lightsail": {"endpoint_prefix": "lightsail", "service_id": "lightsail"},
    "logs": {"endpoint_prefix": "logs", "service_id": "cloudwatch-logs"},
    "machinelearning": {
        "endpoint_prefix": "machinelearning",
        "service_id": "machine-learning",
    },
    "marketplace-entitlement": {
        "endpoint_prefix": "entitlement.marketplace",
        "service_id": "marketplace-entitlement-service",
    },
    "marketplacecommerceanalytics": {
        "endpoint_prefix": "marketplacecommerceanalytics",
        "service_id": "marketplace-commerce-analytics",
    },
    "mediaconvert": {
        "endpoint_prefix": "mediaconvert",
        "service_id": "mediaconvert",
    },
    "medialive": {"endpoint_prefix": "medialive", "service_id": "medialive"},
    "mediapackage": {
        "endpoint_prefix": "mediapackage",
        "service_id": "mediapackage",
    },
    "mediastore": {
        "endpoint_prefix": "mediastore",
        "service_id": "mediastore",
    },
    "mediastore-data": {
        "endpoint_prefix": "data.mediastore",
        "service_id": "mediastore-data",
    },
    "mediatailor": {
        "endpoint_prefix": "api.mediatailor",
        "service_id": "mediatailor",
    },
    "meteringmarketplace": {
        "endpoint_prefix": "metering.marketplace",
        "service_id": "marketplace-metering",
    },
    "mgh": {"endpoint_prefix": "mgh", "service_id": "migration-hub"},
    "mq": {"endpoint_prefix": "mq", "service_id": "mq"},
    "mturk": {"endpoint_prefix": "mturk-requester", "service_id": "mturk"},
    "neptune": {"service_id": "neptune"},
    "opsworks": {"endpoint_prefix": "opsworks", "service_id": "opsworks"},
    "opsworkscm": {
        "endpoint_prefix": "opsworks-cm",
        "service_id": "opsworkscm",
    },
    "organizations": {
        "endpoint_prefix": "organizations",
        "service_id": "organizations",
    },
    "pi": {"endpoint_prefix": "pi", "service_id": "pi"},
    "pinpoint": {"endpoint_prefix": "pinpoint", "service_id": "pinpoint"},
    "polly": {"endpoint_prefix": "polly", "service_id": "polly"},
    "pricing": {"endpoint_prefix": "api.pricing", "service_id": "pricing"},
    "rds": {"endpoint_prefix": "rds", "service_id": "rds"},
    "redshift": {"endpoint_prefix": "redshift", "service_id": "redshift"},
    "rekognition": {
        "endpoint_prefix": "rekognition",
        "service_id": "rekognition",
    },
    "resource-groups": {
        "endpoint_prefix": "resource-groups",
        "service_id": "resource-groups",
    },
    "resourcegroupstaggingapi": {
        "endpoint_prefix": "tagging",
        "service_id": "resource-groups-tagging-api",
    },
    "route53": {"endpoint_prefix": "route53", "service_id": "route-53"},
    "route53domains": {
        "endpoint_prefix": "route53domains",
        "service_id": "route-53-domains",
    },
    "s3": {"endpoint_prefix": "s3", "service_id": "s3"},
    "sagemaker": {
        "endpoint_prefix": "api.sagemaker",
        "service_id": "sagemaker",
    },
    "sagemaker-runtime": {
        "endpoint_prefix": "runtime.sagemaker",
        "service_id": "sagemaker-runtime",
    },
    "sdb": {"endpoint_prefix": "sdb", "service_id": "simpledb"},
    "secretsmanager": {
        "endpoint_prefix": "secretsmanager",
        "service_id": "secrets-manager",
    },
    "serverlessrepo": {
        "endpoint_prefix": "serverlessrepo",
        "service_id": "serverlessapplicationrepository",
    },
    "servicecatalog": {
        "endpoint_prefix": "servicecatalog",
        "service_id": "service-catalog",
    },
    "servicediscovery": {
        "endpoint_prefix": "servicediscovery",
        "service_id": "servicediscovery",
    },
    "ses": {"endpoint_prefix": "email", "service_id": "ses"},
    "shield": {"endpoint_prefix": "shield", "service_id": "shield"},
    "sms": {"endpoint_prefix": "sms", "service_id": "sms"},
    "snowball": {"endpoint_prefix": "snowball", "service_id": "snowball"},
    "sns": {"endpoint_prefix": "sns", "service_id": "sns"},
    "sqs": {"endpoint_prefix": "sqs", "service_id": "sqs"},
    "ssm": {"endpoint_prefix": "ssm", "service_id": "ssm"},
    "stepfunctions": {"endpoint_prefix": "states", "service_id": "sfn"},
    "storagegateway": {
        "endpoint_prefix": "storagegateway",
        "service_id": "storage-gateway",
    },
    "sts": {"endpoint_prefix": "sts", "service_id": "sts"},
    "support": {"endpoint_prefix": "support", "service_id": "support"},
    "swf": {"endpoint_prefix": "swf", "service_id": "swf"},
    "transcribe": {
        "endpoint_prefix": "transcribe",
        "service_id": "transcribe",
    },
    "translate": {"endpoint_prefix": "translate", "service_id": "translate"},
    "waf": {"endpoint_prefix": "waf", "service_id": "waf"},
    "waf-regional": {
        "endpoint_prefix": "waf-regional",
        "service_id": "waf-regional",
    },
    "workdocs": {"endpoint_prefix": "workdocs", "service_id": "workdocs"},
    "workmail": {"endpoint_prefix": "workmail", "service_id": "workmail"},
    "workspaces": {
        "endpoint_prefix": "workspaces",
        "service_id": "workspaces",
    },
    "xray": {"endpoint_prefix": "xray", "service_id": "xray"},
}


def _event_aliases():
    for client_name in SERVICES.keys():
        service_id = SERVICES[client_name]['service_id']
        yield client_name, service_id


def _event_aliases_with_endpoint_prefix():
    for client_name in SERVICES.keys():
        endpoint_prefix = SERVICES[client_name].get('endpoint_prefix')
        if endpoint_prefix is not None:
            yield client_name, endpoint_prefix


@pytest.mark.parametrize(
    "client_name, endpoint_prefix", _event_aliases_with_endpoint_prefix()
)
def test_event_alias_by_endpoint_prefix(client_name, endpoint_prefix):
    _assert_handler_called(client_name, endpoint_prefix)


@pytest.mark.parametrize("client_name, service_id", _event_aliases())
def test_event_alias_by_service_id(client_name, service_id):
    _assert_handler_called(client_name, service_id)


@pytest.mark.parametrize("client_name, service_id", _event_aliases())
def test_event_alias_by_client_name(client_name, service_id):
    _assert_handler_called(client_name, client_name)


def _assert_handler_called(client_name, event_part):
    hook_calls = []

    def _hook(**kwargs):
        hook_calls.append(kwargs['event_name'])

    session = _get_session()
    session.register(f'creating-client-class.{event_part}', _hook)
    session.create_client(client_name)
    assert len(hook_calls) == 1


def _get_session():
    session = Session()
    session.set_credentials('foo', 'bar')
    session.set_config_variable('region', 'us-west-2')
    session.config_filename = 'no-exist-foo'
    return session



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from tests import BaseSessionTest, mock


class RecordingHandler:
    def __init__(self):
        self.recorded_events = []

    def record(self, event_name, **kwargs):
        self.recorded_events.append((event_name, kwargs))


class TestClientEvents(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.region = 'us-west-2'
        self.client = self.session.create_client('ec2', self.region)

    def test_emit_response_received(self):
        recording_handler = RecordingHandler()
        self.client.meta.events.register(
            'response-received.ec2.DescribeRegions', recording_handler.record
        )
        with mock.patch(
            'botocore.httpsession.URLLib3Session.send'
        ) as mock_send:
            response_body = (
                b'<?xml version="1.0" ?>'
                b'<DescribeRegionsResponse xmlns="">'
                b'</DescribeRegionsResponse>'
            )
            mock_send.return_value = mock.Mock(
                status_code=200, headers={}, content=response_body
            )
            self.client.describe_regions()
        self.assertEqual(
            recording_handler.recorded_events,
            [
                (
                    'response-received.ec2.DescribeRegions',
                    {
                        'exception': None,
                        'response_dict': {
                            'body': response_body,
                            'headers': {},
                            'context': mock.ANY,
                            'status_code': 200,
                        },
                        'parsed_response': {'ResponseMetadata': mock.ANY},
                        'context': mock.ANY,
                    },
                )
            ],
        )

    def test_emit_response_received_for_exception(self):
        recording_handler = RecordingHandler()
        self.client.meta.events.register(
            'response-received.ec2.DescribeRegions', recording_handler.record
        )
        with mock.patch(
            'botocore.httpsession.URLLib3Session.send'
        ) as mock_send:
            raised_exception = RuntimeError('Unexpected exception')
            mock_send.side_effect = raised_exception
            with self.assertRaises(RuntimeError):
                self.client.describe_regions()
        self.assertEqual(
            recording_handler.recorded_events,
            [
                (
                    'response-received.ec2.DescribeRegions',
                    {
                        'exception': raised_exception,
                        'response_dict': None,
                        'parsed_response': None,
                        'context': mock.ANY,
                    },
                )
            ],
        )



import json

import pytest

from botocore.config import Config
from botocore.exceptions import InvalidEndpointConfigurationError
from tests import BaseSessionTest, ClientHTTPStubber, requires_crt


class TestClientEvents(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.region = "us-east-1"

    def create_eventbridge_client(self, region=None, **kwargs):
        if region is None:
            region = self.region
        client = self.session.create_client("events", region, **kwargs)
        return client

    def create_stubbed_eventbridge_client(
        self, with_default_responses=False, **kwargs
    ):
        client = self.create_eventbridge_client(**kwargs)
        http_stubber = ClientHTTPStubber(client)
        http_stubber.start()
        if with_default_responses:
            http_stubber.add_response()
            http_stubber.add_response()
        return client, http_stubber

    def _default_put_events_args(self):
        return {
            "Entries": [
                {
                    "Source": "test",
                    "Resources": [
                        "resource",
                    ],
                    "DetailType": "my-detail",
                    "Detail": "detail",
                    "EventBusName": "my-bus",
                },
            ]
        }

    def _assert_multi_region_endpoint(self, request, endpoint_id, suffix=None):
        if suffix is None:
            suffix = "amazonaws.com"
        assert (
            request.url == f"https://{endpoint_id}.endpoint.events.{suffix}/"
        )

    def _assert_sigv4a_headers(self, request):
        assert request.headers["x-amz-region-set"] == b"*"
        assert request.headers["authorization"].startswith(
            b"AWS4-ECDSA-P256-SHA256 Credential="
        )

    def _assert_params_in_body(self, request, params):
        assert len(params) > 0
        body = json.loads(request.body)
        for key, value in params:
            assert body[key] == value

    def test_put_event_default_endpoint(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
        )
        with stubber:
            client.put_events(**self._default_put_events_args())
        assert (
            stubber.requests[0].url
            == "https://events.us-east-1.amazonaws.com/"
        )
        assert b"EndpointId" not in stubber.requests[0].body

    def test_put_event_default_endpoint_explicit_configs(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
            config=Config(
                use_dualstack_endpoint=False,
                use_fips_endpoint=False,
            ),
        )
        with stubber:
            client.put_events(**self._default_put_events_args())
        assert (
            stubber.requests[0].url
            == "https://events.us-east-1.amazonaws.com/"
        )
        assert b"EndpointId" not in stubber.requests[0].body

    @requires_crt()
    def test_put_event_endpoint_id(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
        )
        default_args = self._default_put_events_args()
        endpoint_id = "abc123.456def"

        with stubber:
            client.put_events(EndpointId=endpoint_id, **default_args)

        self._assert_params_in_body(
            stubber.requests[0],
            [
                ("EndpointId", endpoint_id),
            ],
        )
        self._assert_multi_region_endpoint(stubber.requests[0], endpoint_id)
        self._assert_sigv4a_headers(stubber.requests[0])

    @requires_crt()
    def test_put_event_endpoint_id_explicit_config(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
            config=Config(
                use_dualstack_endpoint=False,
                use_fips_endpoint=False,
            ),
        )
        default_args = self._default_put_events_args()
        endpoint_id = "abc123.456def"

        with stubber:
            client.put_events(EndpointId=endpoint_id, **default_args)

        self._assert_params_in_body(
            stubber.requests[0],
            [
                ("EndpointId", endpoint_id),
            ],
        )
        self._assert_multi_region_endpoint(stubber.requests[0], endpoint_id)
        self._assert_sigv4a_headers(stubber.requests[0])

    @requires_crt()
    def test_put_event_bad_endpoint_id(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
        )
        default_args = self._default_put_events_args()
        endpoint_id = "badactor.com?foo=bar"

        with pytest.raises(InvalidEndpointConfigurationError):
            client.put_events(EndpointId=endpoint_id, **default_args)

    @requires_crt()
    def test_put_event_bad_endpoint_id_explicit_config(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
            config=Config(
                use_dualstack_endpoint=False,
                use_fips_endpoint=False,
            ),
        )
        default_args = self._default_put_events_args()
        endpoint_id = "badactor.com?foo=bar"

        with pytest.raises(InvalidEndpointConfigurationError):
            client.put_events(EndpointId=endpoint_id, **default_args)

    @requires_crt()
    def test_put_event_empty_endpoint_id(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
        )
        default_args = self._default_put_events_args()
        endpoint_id = ""

        with pytest.raises(InvalidEndpointConfigurationError):
            client.put_events(EndpointId=endpoint_id, **default_args)

    @requires_crt()
    def test_put_event_empty_endpoint_id_explicit_config(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
            config=Config(
                use_dualstack_endpoint=False,
                use_fips_endpoint=False,
            ),
        )
        default_args = self._default_put_events_args()
        endpoint_id = ""

        with pytest.raises(InvalidEndpointConfigurationError):
            client.put_events(EndpointId=endpoint_id, **default_args)

    def test_put_event_default_dualstack_endpoint(self):
        config = Config(use_dualstack_endpoint=True, use_fips_endpoint=False)
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True, config=config
        )
        default_args = self._default_put_events_args()

        with stubber:
            client.put_events(**default_args)
        assert stubber.requests[0].url == "https://events.us-east-1.api.aws/"

    @requires_crt()
    def test_put_events_endpoint_id_dualstack(self):
        config = Config(use_dualstack_endpoint=True, use_fips_endpoint=False)
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True, config=config
        )
        default_args = self._default_put_events_args()
        endpoint_id = "abc123.456def"

        with stubber:
            client.put_events(EndpointId=endpoint_id, **default_args)

        self._assert_params_in_body(
            stubber.requests[0],
            [
                ("EndpointId", endpoint_id),
            ],
        )
        self._assert_multi_region_endpoint(
            stubber.requests[0], endpoint_id, suffix="api.aws"
        )
        self._assert_sigv4a_headers(stubber.requests[0])

    def test_put_events_default_fips_endpoint(self):
        config = Config(use_dualstack_endpoint=False, use_fips_endpoint=True)
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True, config=config
        )
        default_args = self._default_put_events_args()

        with stubber:
            client.put_events(**default_args)
        assert (
            stubber.requests[0].url
            == "https://events-fips.us-east-1.amazonaws.com/"
        )

    @requires_crt()
    def test_put_events_endpoint_id_fips(self):
        config = Config(use_dualstack_endpoint=False, use_fips_endpoint=True)
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True, config=config
        )
        default_args = self._default_put_events_args()
        endpoint_id = "abc123.456def"

        with pytest.raises(InvalidEndpointConfigurationError):
            client.put_events(EndpointId=endpoint_id, **default_args)

    def test_put_events_default_dualstack_fips_endpoint(self):
        config = Config(use_dualstack_endpoint=True, use_fips_endpoint=True)
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True, config=config
        )
        default_args = self._default_put_events_args()

        with stubber:
            client.put_events(**default_args)
        assert (
            stubber.requests[0].url == "https://events-fips.us-east-1.api.aws/"
        )

    @requires_crt()
    def test_put_events_endpoint_id_dualstack_fips(self):
        config = Config(use_dualstack_endpoint=True, use_fips_endpoint=True)
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True, config=config
        )
        default_args = self._default_put_events_args()
        endpoint_id = "abc123.456def"

        with pytest.raises(InvalidEndpointConfigurationError) as e:
            client.put_events(EndpointId=endpoint_id, **default_args)
        assert (
            "FIPS is not supported with EventBridge multi-region endpoints"
            in str(e.value)
        )

    def test_put_events_default_gov_endpoint(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
            region="us-iso-east-1",
        )
        default_args = self._default_put_events_args()

        with stubber:
            client.put_events(**default_args)
        assert (
            stubber.requests[0].url
            == "https://events.us-iso-east-1.c2s.ic.gov/"
        )

    @requires_crt()
    def test_put_events_endpoint_id_gov(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True,
            region="us-iso-east-1",
        )
        default_args = self._default_put_events_args()
        endpoint_id = "abc123.456def"

        with stubber:
            client.put_events(EndpointId=endpoint_id, **default_args)

        self._assert_params_in_body(
            stubber.requests[0],
            [
                ("EndpointId", endpoint_id),
            ],
        )
        self._assert_multi_region_endpoint(
            stubber.requests[0], endpoint_id, suffix="c2s.ic.gov"
        )
        self._assert_sigv4a_headers(stubber.requests[0])

    def test_put_events_default_custom_endpoint(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True, endpoint_url="https://example.org"
        )
        default_args = self._default_put_events_args()

        with stubber:
            client.put_events(**default_args)
        assert stubber.requests[0].url == "https://example.org/"

    @requires_crt()
    def test_put_events_endpoint_id_custom(self):
        client, stubber = self.create_stubbed_eventbridge_client(
            with_default_responses=True, endpoint_url="https://example.org"
        )
        default_args = self._default_put_events_args()
        endpoint_id = "abc123.456def"

        with stubber:
            client.put_events(EndpointId=endpoint_id, **default_args)

        self._assert_params_in_body(
            stubber.requests[0],
            [
                ("EndpointId", endpoint_id),
            ],
        )
        assert stubber.requests[0].url == "https://example.org/"
        self._assert_sigv4a_headers(stubber.requests[0])
